<?php

namespace Drupal\Tests\popular_tags\FunctionalJavascript;

use Drupal\Core\Database\Database;
use Drupal\Tests\taxonomy\Traits\TaxonomyTestTrait;
use Drupal\Tests\field\Traits\EntityReferenceTestTrait;
use Drupal\FunctionalJavascriptTests\WebDriverTestBase;

/**
 * Base class for testing the functionality of popular tags.
 *
 * @group popular_tags
 */
class PopularTagsEditTagsTest extends WebDriverTestBase {

  use TaxonomyTestTrait;
  use EntityReferenceTestTrait;

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  protected $limit = 5;
  protected $fieldName;

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['node', 'user', 'taxonomy', 'popular_tags'];

  protected function setUp(): void {
    parent::setUp();

    $this->createContentType([
      'type' => 'ptag_page',
      'name' => 'Popular Tag Page'
    ]);

    $vocabulary = $this->createVocabulary();
    $tids = [];
    for($i = 0; $i < 10; $i++) {
      $tids[] = $this->createTerm($vocabulary)->id();
    }

    $fieldName = 'field_' . $vocabulary->id();
    $this->fieldName = $fieldName;

    $handler_settings = [
      'target_bundles' => [
        $vocabulary->id() => $vocabulary->id(),
      ],
      'auto_create' => TRUE,
    ];
    $this->createEntityReferenceField('node', 'ptag_page', $fieldName, 'Tags', 'taxonomy_term', 'default', $handler_settings, 5);

    $display_repository = \Drupal::service('entity_display.repository');
    $display_repository->getFormDisplay('node', 'ptag_page')
      ->setComponent($fieldName, [
        'type' => 'entity_reference_autocomplete_tags',
        'third_party_settings' => [
          'popular_tags' => [
            'popular_tags' => [
              'use' => TRUE,
              'limit' => $this->limit,
              'showhide' => FALSE
            ],
          ],
        ],
      ])
      ->save();

    $authorUser = $this->createUser(['edit own ptag_page content']);

    for($i = 0; $i < 10; $i++) {
      $node = $this->createNode([
        'uid' => $authorUser->id(),
        'type' => 'ptag_page',
      ]);
      $rands = $this->generateRandomNumbers();
      $node->set($fieldName, [$tids[0], $tids[$rands[0]], $tids[$rands[1]]]);
      $node->save();
      unset($node);
    }

    $this->drupalLogin($authorUser);

  }

  public function testIsShowingMostPopularTerms() {
    // $this->createScreenshot('/tmp/screenshot.jpg');

    $this->drupalGet('/node/1/edit');

    $session_assert = $this->assertSession();
    $session_assert->fieldExists("{$this->fieldName}[target_id]");
    $element = $session_assert->elementExists('css', '.tag-terms');

    $terms = $this->getMostUsedTerms();
    $tags = $this->getSession()->getPage()->findAll('css', '.tag-terms > a.term');
    $this->assertTrue(!empty($tags) && count($tags) == $this->limit);
    // Make sure that terms are displayed in the correct order
    // and have the correct "frequency" in their data-usage attribute.
    foreach($tags as $idx => $tag) {
      $this->assertTrue($tag->getText() == $terms[$idx]->name, 'No Match');
      $this->assertTrue($tag->getAttribute('data-usage') == "{$terms[$idx]->count}", 'Counts Don\'t match');
    }
    // Make sure that a term not within the limit isn't displayed.
    $this->assertTrue(strpos($element->getText(), $terms[$this->limit + 1]->name) === FALSE);
  }

  public function testCanAddNewTerm() {

    $nid = mt_rand(1, 9);
    $this->drupalGet("/node/$nid/edit");

    $field = $this->getSession()->getPage()->find('css', "input[name='{$this->fieldName}[target_id]']");
    $initialValue = $field->getValue();
    $tagAdded = '';
    $tags = $this->getSession()->getPage()->findAll('css', '.tag-terms > a.term');
    foreach($tags as $tag) {
      // Find a term that hasn't been added yet.
      $tagText = $tag->getText();
      if (strpos($initialValue, $tagText) === FALSE) {
        $tagAdded = $tagText;
        $tag->click();
        sleep(2);
        break;
      }
    }
    $currentValue = $field->getValue();
    // Make sure the field's value has changed.
    $this->assertTrue($currentValue !== $initialValue);
    // Make sure the field's value includes the term we attempted to add.
    $this->assertTrue(strpos($currentValue, $tagAdded) !== FALSE);
  }

  public function testCannotAddExistingTerm() {

    $nid = mt_rand(1, 9);
    $this->drupalGet("/node/$nid/edit");

    $field = $this->getSession()->getPage()->find('css', "input[name='{$this->fieldName}[target_id]']");
    $initialValue = $field->getValue();
    $tag = $this->getSession()->getPage()->find('css', '.tag-terms > a.term');
    $tag->click();
    sleep(2);
    $this->assertTrue($field->getValue() === $initialValue);
  }

  // Helper function to find the most commonly used terms and their frequencies.
  protected function getMostUsedTerms() {
    $query = Database::getConnection()->select('taxonomy_index', 'ti');
    $query->leftJoin('taxonomy_term_field_data', 'ttfd', 'ttfd.tid = ti.tid');
    $query->fields('ttfd', ['name']);
    $query->addExpression('count(*)', 'count');
    $query->addExpression("GROUP_CONCAT(ti.nid SEPARATOR ',')", 'nodes');
    $query->groupBy('ti.tid')
      ->orderBy('count', 'DESC')
      ->orderBy('ttfd.name', 'ASC');
    $results = $query->execute()->fetchAll();
    return !empty($results) ? $results : [];
  }

  // Helper function to generate random term ids.
  protected function generateRandomNumbers() {
    $nums = [];
    for($i = 0; $i < 2; $i++) {
      $isDone = FALSE;
      while($isDone === FALSE) {
        // We create 10 terms and pick a random number between 1 and 9
        // Skip 0 since all nodes are tagged with that term.
        $rand = mt_rand(1, 9);
        if (!in_array($rand, $nums)) {
          $nums[] = $rand;
          $isDone = TRUE;
        }
      }
    }
    return $nums;
  }

}